// Copyright (C) 2005, Ondra Kamenik

// $Id: fine_atoms.cpp 762 2006-05-22 13:00:07Z kamenik $

#include "utils/cc/exception.h"

#include "parser_exception.h"
#include "fine_atoms.h"

using namespace ogp;

FineAtoms::FineAtoms(const FineAtoms& fa)
	: DynamicAtoms(fa), params(), endovars(), exovars(),
	  endo_order(NULL), exo_order(NULL), der_atoms(fa.der_atoms),
	  endo_atoms_map(fa.endo_atoms_map),
	  exo_atoms_map(fa.exo_atoms_map)
{
	// fill in params
	for (unsigned int i = 0; i < fa.params.size(); i++) {
		const char* s = varnames.query(fa.params[i]);
		if (! s)
			throw ogu::Exception(__FILE__, __LINE__,
								 "Parameter name does not exist in FineAtoms copy cosntructor");
		params.push_back(s);
	}
	// fill in endovars
	for (unsigned int i = 0; i < fa.endovars.size(); i++) {
		const char* s = varnames.query(fa.endovars[i]);
		if (! s)
			throw ogu::Exception(__FILE__, __LINE__,
								 "Endo variable name does not exist in FineAtoms copy cosntructor");
		endovars.push_back(s);
	}
	// fill in exovars
	for (unsigned int i = 0; i < fa.exovars.size(); i++) {
		const char* s = varnames.query(fa.exovars[i]);
		if (! s)
			throw ogu::Exception(__FILE__, __LINE__,
								 "Exo variable name does not exist in FineAtoms copy cosntructor");
		exovars.push_back(s);
	}

	if (fa.endo_order)
		endo_order = fa.endo_order->clone(endovars, *this);

	if (fa.exo_order)
		exo_order = fa.exo_order->clone(exovars, *this);
}

int FineAtoms::check_variable(const char* name) const
{
	string str;
	int ll;
	parse_variable(name, str, ll);
	if (varnames.query(str.c_str()))
		return DynamicAtoms::check_variable(name);
	else {
		throw ParserException(string("Variable <")+str+"> not declared.",0);
		return -1;
	}
}

int FineAtoms::num_exo_periods() const
{
	int mlead, mlag;
	exovarspan(mlead, mlag);
	return mlead-mlag+1;
}

void FineAtoms::parsing_finished(VarOrdering::ord_type ot)
{
	int mlead, mlag;
	endovarspan(mlead, mlag);
	if (mlag < -1 || mlead > 1)
		throw ogu::Exception(__FILE__,__LINE__,
							 "Wrong minlag/maxlead in FineAtoms::parsing_finished");

	// create endo_order and exo_order
	if (endo_order)
		delete endo_order;
	if (exo_order)
		delete exo_order;
	if (ot == VarOrdering::pbspbfbf)
		endo_order = new EndoVarOrdering1(endovars, *this);
	else
		endo_order = new EndoVarOrdering2(endovars, *this);
	endo_order->do_ordering();
	exo_order = new ExoVarOrdering(exovars, *this);
	exo_order->do_ordering();

	// concatenate der atoms from endo_order and exo_order
	der_atoms.clear();
	der_atoms.insert(der_atoms.end(), 
					 endo_order->get_der_atoms().begin(),
					 endo_order->get_der_atoms().end());
	der_atoms.insert(der_atoms.end(), 
					 exo_order->get_der_atoms().begin(),
					 exo_order->get_der_atoms().end());

	// create endo_atoms_map; der_atoms is a concatenation, so it is easy
	int endo_atoms = endo_order->get_der_atoms().size();
	for (int i = 0; i < endo_atoms; i++)
		endo_atoms_map.push_back(i);
	// create exo_atoms_map
	int exo_atoms = exo_order->get_der_atoms().size();
	for (int i = 0; i < exo_atoms; i++)
		exo_atoms_map.push_back(endo_atoms + i);
}

vector<int> FineAtoms::variables() const
{
	if (endo_order) {
		return der_atoms;
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::variables called before parsing_finished");
		return vector<int>();
	}
}

int FineAtoms::nstat() const
{
	if (endo_order) {
		return endo_order->nstat();
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::nstat called before parsing_finished");
		return -1;
	}
}

int FineAtoms::npred() const
{
	if (endo_order) {
		return endo_order->npred();
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::npred called before parsing_finished");
		return -1;
	}
}

int FineAtoms::nboth() const
{
	if (endo_order) {
		return endo_order->nboth();
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::nboth called before parsing_finished");
		return -1;
	}
}

int FineAtoms::nforw() const
{
	if (endo_order) {
		return endo_order->nforw();
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::nforw called before parsing_finished");
		return -1;
	}
}

int FineAtoms::get_pos_of_endo(int t) const
{
	if (endo_order) {
		return endo_order->get_pos_of(t);
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::get_pos_of_endo called before parsing_finished");
		return -1;
	}
}

int FineAtoms::get_pos_of_exo(int t) const
{
	if (exo_order) {
		return exo_order->get_pos_of(t);
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::get_pos_of_exo called before parsing_finished");
		return -1;
	}
}

int FineAtoms::get_pos_of_all(int t) const
{
	if (endo_order && exo_order) {
		if (endo_order->check(t))
			return endo_order->get_pos_of(t);
		else if (exo_order->check(t))
			return endo_order->length() + exo_order->get_pos_of(t);
		else {
			throw ogu::Exception(__FILE__,__LINE__,
								 "Atom is not endo nor exo in FineAtoms::get_pos_of_all");
			return -1;
		}
	} else {
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::get_pos_of_exo called before parsing_finished");
		return -1;
	}
}

const vector<int>& FineAtoms::y2outer_endo() const
{
	if (! endo_order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::y2outer_endo called before parsing_finished");
	return endo_order->get_y2outer();
}

const vector<int>& FineAtoms::outer2y_endo() const
{
	if (! endo_order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::outer2y_endo called before parsing_finished");
	return endo_order->get_outer2y();
}

const vector<int>& FineAtoms::y2outer_exo() const
{
	if (! exo_order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::y2outer_endo called before parsing_finished");
	return exo_order->get_y2outer();
}

const vector<int>& FineAtoms::outer2y_exo() const
{
	if (! exo_order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::outer2y_exo called before parsing_finished");
	return exo_order->get_outer2y();
}

const vector<int>& FineAtoms::get_endo_atoms_map() const
{
	if (! endo_order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::get_endo_atoms_map called before parsing_finished");
	return endo_atoms_map;
}

const vector<int>& FineAtoms::get_exo_atoms_map() const
{
	if (! exo_order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "FineAtoms::get_exo_atoms_map called before parsing_finished");
	return exo_atoms_map;
}
