// Copyright (C) 2005, Ondra Kamenik

// $Id: fine_atoms.h 762 2006-05-22 13:00:07Z kamenik $

#ifndef OGP_FINE_ATOMS_H
#define OGP_FINE_ATOMS_H

#include "dynamic_atoms.h"

#include <vector>
#include <string>

namespace ogp {

	using std::vector;
	using std::string;

	/** This is just ordering used for endogenous variables. It
	 * assumes that we have only time t-1, t, and t+1, orders them as
	 * pred(t-1), both(t-1), stat(t), pred(t), both(t), forw(t),
	 * both(t+1), forw(t+1). */
	class EndoVarOrdering1 : public VarOrdering {
	public:
		EndoVarOrdering1(const vector<const char*>& vnames, const DynamicAtoms& a)
			: VarOrdering(vnames, a) {}
		EndoVarOrdering1(const EndoVarOrdering1& vo, const vector<const char*>& vnames,
						 const DynamicAtoms& a)
			: VarOrdering(vo, vnames, a) {}
		VarOrdering* clone(const vector<const char*>& vnames, const DynamicAtoms& a) const
			{return new EndoVarOrdering1(*this, vnames, a);}
		void do_ordering()
			{do_pbspbfbf();}
	};

	/** This is just another ordering used for endogenous
	 * variables. It assumes that we have only time t-1, t, and t+1,
	 * orders them as both(t+1), forw(t+1), pred(t-1), both(t-1),
	 * stat(t), pred(t), both(t), forw(t). */
	class EndoVarOrdering2 : public VarOrdering {
	public:
		EndoVarOrdering2(const vector<const char*>& vnames, const DynamicAtoms& a)
			: VarOrdering(vnames, a) {}
		EndoVarOrdering2(const EndoVarOrdering2& vo, const vector<const char*>& vnames,
						 const DynamicAtoms& a)
			: VarOrdering(vo, vnames, a) {}
		VarOrdering* clone(const vector<const char*>& vnames, const DynamicAtoms& a) const
			{return new EndoVarOrdering2(*this, vnames, a);}
		void do_ordering()
			{do_bfspbfpb();}
	};

	/** This is just ordering used for exogenous variables. It makes
	 * no assumptions about their timing. It orders them from the
	 * least time to the latest time. */
	class ExoVarOrdering : public VarOrdering {
	public:
		ExoVarOrdering(const vector<const char*>& vnames, const DynamicAtoms& a)
			: VarOrdering(vnames, a) {}
		ExoVarOrdering(const ExoVarOrdering& vo, const vector<const char*>& vnames,
					   const DynamicAtoms& a)
			: VarOrdering(vo, vnames, a) {}
		VarOrdering* clone(const vector<const char*>& vnames, const DynamicAtoms& a) const
			{return new ExoVarOrdering(*this, vnames, a);}
		void do_ordering()
			{do_increasing_time();}
	};

	class FineAtoms : public DynamicAtoms {
	protected:
		/** The vector of parameters names. The order gives the order
		 * the data is communicated with outside world. */
		vector<const char*> params;
		/** The vector of endogenous variables. This defines the order
		 * like parameters. */
		vector<const char*> endovars;
		/** The vector of exogenous variables. Also defines the order
		 * like parameters and endovars. */
		vector<const char*> exovars;

		/** This is the ordering of all atoms corresponding to
		 * endogenous variables. It is constructed by parsing_finished()
		 * method, which should be called after all parsing jobs have
		 * been finished. */ 
		VarOrdering* endo_order;
		/** This is the ordering of all atoms corresponding to
		 * exogenous variables. It has the same handling as
		 * endo_order. */
		VarOrdering* exo_order;
		/** This vector defines a set of atoms as tree indices used
		 * for differentiation. The order of the atoms in this vector
		 * defines ordering of the derivative tensors. The ordering is
		 * a concatenation of atoms from endo_order and then
		 * exo_order. This vector is setup by parsing_finished() and
		 * is returned by variables(). */
		vector<int> der_atoms;
		/** This is a mapping from endogenous atoms to all atoms in
		 * der_atoms member. The mapping maps index in endogenous atom
		 * ordering to index (not value) in der_atoms. It is useful if
		 * one wants to evaluate derivatives wrt only endogenous
		 * variables. It is set by parsing_finished(). By definition,
		 * it is monotone. */
		vector<int> endo_atoms_map;
		/** This is a mapping from exogenous atoms to all atoms in
		 * der_atoms member. It is the same as endo_atoms_map for
		 * atoms of exogenous variables. */
		vector<int> exo_atoms_map;
	public:
		FineAtoms()
			: endo_order(NULL), exo_order(NULL) {}
		FineAtoms(const FineAtoms& fa);
		/** Deletes endo_order and exo_order. */
		virtual ~FineAtoms()
			{
				if (endo_order) delete endo_order;
				if (exo_order) delete exo_order;
			}
		/** Overrides DynamicAtoms::check_variable so that the error
		 * would be raised if the variable name is not declared. A
		 * variable is declared by inserting it to
		 * DynamicAtoms::varnames. This is a responsibility of a
		 * subclass. */
		int check_variable(const char* name) const;
		/** This calculates min lag and max lead of endogenous variables. */
		void endovarspan(int& mlead, int& mlag) const
			{varspan(endovars, mlead, mlag);}
		/** This calculates mim lag and max lead of exogenous variables. */
		void exovarspan(int& mlead, int& mlag) const
			{varspan(exovars, mlead, mlag);}
		/** This calculates the number of periods in which at least
		 * one exogenous variable occurs. */
		int num_exo_periods() const;
		const vector<const char*>& get_params() const
			{return params;}
		const vector<const char*>& get_endovars() const
			{return endovars;}
		const vector<const char*>& get_exovars() const
			{return exovars;}
		/** This constructs solv_ordering and makes the indices
		 * returned by variables method available. In addition, it
		 * makes nstat, npred, nboth, nforw available. */
		void parsing_finished(VarOrdering::ord_type ot);
		/** Return the atoms with respect which we are going to
		 * differentiate. This must be called after
		 * parsing_finished. */
		vector<int> variables() const;
		/** Return the number of static. */
		int nstat() const;
		/** Return the number of predetermined. */
		int npred() const;
		/** Return the number of both. */
		int nboth() const;
		/** Return the number of forward looking. */
		int nforw() const;
		/** Return the index of an endogenous atom given by tree index in
		 * the endo ordering. This must be also called only after
		 * parsing_finished(). */
		int get_pos_of_endo(int t) const;
		/** Return the index of an exogenous atom given by tree index in
		 * the exo ordering. This must be also called only after
		 * parsing_finished(). */
		int get_pos_of_exo(int t) const;
		/** Return the index of either endogenous or exogenous atom
		 * given by tree index in the concatenated ordering of
		 * endogenous and exogenous atoms. This must be also called
		 * only after parsing_finished(). */
		int get_pos_of_all(int t) const;
		/** Return the mapping from endogenous at time t to outer
		 * ordering of endogenous. */
		const vector<int>& y2outer_endo() const;
		/** Return the mapping from the outer ordering of endogenous to endogenous
		 * at time t. */
		const vector<int>& outer2y_endo() const;
		/** Return the mapping from exogenous at time t to outer
		 * ordering of exogenous. */
		const vector<int>& y2outer_exo() const;
		/** Return the mapping from the outer ordering of exogenous to exogenous
		 * at time t. */
		const vector<int>& outer2y_exo() const;
		/** Return the endo_atoms_map. */
		const vector<int>& get_endo_atoms_map() const;
		/** Return the exo_atoms_map. */
		const vector<int>& get_exo_atoms_map() const;
		/** Return the number of endogenous variables at time t-1, these are state
		 * variables. */
		int nys() const
			{return npred()+nboth();}
		/** Return the number of endogenous variables at time t+1. */
		int nyss() const
			{return nboth()+nforw();}
		/** Return the number of endogenous variables. */
		int ny() const
			{return nstat()+npred()+nboth()+nforw();}
		/** Return the number of exogenous variables. */
		int nexo() const
			{return (int)exovars.size();}
		/** Return the number of parameters. */
		int np() const
			{return (int)(params.size());}
	};

};

#endif

// Local Variables:
// mode:C++
// End:
