// Copyright (C) 2005, Ondra Kamenik

// $Id: formula_parser.cpp 762 2006-05-22 13:00:07Z kamenik $

#include "utils/cc/pascal_triangle.h"
#include "utils/cc/exception.h"

#include "parser_exception.h"
#include "location.h"
#include "formula_parser.h"
#include "formula_tab.hh"

using namespace ogp;

extern location_type fmla_lloc;

FormulaParser::FormulaParser(const FormulaParser& fp, Atoms& a)
	: otree(fp.otree), atoms(a), formulas(fp.formulas), ders()
{
	// create derivatives
	for (unsigned int i = 0; i < fp.ders.size(); i++)
		ders.push_back(new FormulaDerivatives(*(fp.ders[i])));
}

FormulaParser::~FormulaParser()
{
	while (ders.size() > 0) {
		delete ders.back();
		ders.pop_back();
	}
}

void FormulaParser::differentiate(int max_order)
{
	vector<int> vars;
	vars = atoms.variables();
	for (unsigned int i = 0; i < formulas.size(); i++)
		ders.push_back(new FormulaDerivatives(otree, vars, formulas[i], max_order));
}

const FormulaDerivatives& FormulaParser::derivatives(int i) const
{
	if (i < (int)ders.size())
		return *(ders[i]);
	else
		throw ogu::Exception(__FILE__,__LINE__,
							 "Wrong formula index in FormulaParser::derivatives");
	return *(ders[0]); // just because of compiler
}



void FormulaParser::add_formula(int t)
{
	formulas.push_back(t);
}

int FormulaParser::add_binary(code_t code, int t1, int t2)
{
	return otree.add_binary(code, t1, t2);
}

int FormulaParser::add_unary(code_t code, int t)
{
	return otree.add_unary(code, t);
}

int FormulaParser::add_nulary(const char* str)
{
	int t = -1;
	try {
		t = atoms.check(str);
	} catch (const ParserException& e) {
		throw ParserException(e, fmla_lloc.off);
	}
	if (t == -1) {
		t = otree.add_nulary();
		atoms.assign(str, t);
	}
	return t;
}

/** Global symbols for passing info to parser. */
FormulaParser* fparser;

/** The declarations of functions defined in formula_ll.cc and
 * formula_tab.cc generated from formula.lex and formula.y */
void* fmla__scan_buffer(char*, unsigned int);
void fmla__destroy_buffer(void*);
void fmla_parse();
extern location_type fmla_lloc; 

/** This makes own copy of provided data, sets the buffer for the
 * parser with fmla_scan_buffer, and launches fmla_parse(). Note that
 * the pointer returned from fmla_scan_buffer must be freed at the
 * end. */ 
void FormulaParser::parse(int length, const char* stream)
{
	char* buffer = new char[length+2];
	strncpy(buffer, stream, length);
	buffer[length] = '\0';
	buffer[length+1] = '\0';
	fmla_lloc.off = 0;
	fmla_lloc.ll = 0;
	void* p = fmla__scan_buffer(buffer, (unsigned int)length+2);
	fparser = this;
	fmla_parse();
	delete [] buffer;
	fmla__destroy_buffer(p);
}

void FormulaParser::error(const char* mes) const
{
	throw ParserException(mes, fmla_lloc.off);
}

int FormulaParser::last_formula() const
{
	int res = -1;
	for (unsigned int i = 0; i < formulas.size(); i++)
		if (res < formulas[i])
			res = formulas[i];
	return res;
}

void FormulaParser::print() const
{
	atoms.print();
	for (unsigned int i = 0; i < formulas.size(); i++) {
		printf("formula %d:\n", formulas[i]);
		otree.print_operation(formulas[i]);
	}
	for (unsigned int i = 0; i < ders.size(); i++) {
		printf("derivatives for the formula %d:\n", formulas[i]);
		ders[i]->print(otree);
	}
}

/** This constructor makes a vector of indices for formulas
 * corresponding to derivatives of the given formula. The formula is
 * supposed to belong to the provided tree, the created derivatives
 * are added to the tree. 
 *
 * The algorithm is as follows. First, the formula f is stored as
 * d[0], since it is zero-th derivative. Then we cycle through orders
 * from 1 to the maximum order. At each order, the derivatives are
 * obtained by differentiation of formulas from previous order with
 * respect to all variables. So next inner loop is through all
 * variables in the ordering provided by argument vars. Each variable
 * is not used for differentiation of all formulas (if so we would get
 * the unfolded scheme with repeating symmetric derivatives), but only
 * with formulas corresponding to folded derivatives of the lower
 * order (k-1) and variables from the variable up to the end (nvar -
 * ivar). So this number is noverk(nvar-ivar + k-1 -1, k-1).
*/
FormulaDerivatives::FormulaDerivatives(OperationTree& otree,
									   const vector<int>& vars, int f, int max_order)
	: d(new int[ptriang.noverk(vars.size()+max_order, max_order)]),
	  nvar(vars.size()), order(max_order)
{
	d[0] = f;
	int cnt = 1;
	for (int k = 1; k <= order; k++) {
		int last_order_end = cnt;
		for (int ivar = 0; ivar < nvar; ivar++) {
			int last_order_beg = last_order_end - ptriang.noverk(nvar-ivar+k-1-1,k-1);
			for (int ff = last_order_beg; ff < last_order_end; ff++) {
				int der = otree.add_derivative(d[ff], vars[ivar]);
				d[cnt++] = der;
			}
		}
	}
}

FormulaDerivatives::FormulaDerivatives(const FormulaDerivatives& fd)
	: d(new int[ptriang.noverk(fd.nvar+fd.order, fd.order)]),
	  nvar(fd.nvar), order(fd.order)
{
	memcpy(d, fd.d, ptriang.noverk(fd.nvar+fd.order, fd.order)*sizeof(int));
}

int FormulaDerivatives::derivative(const FoldMultiIndex& mi) const
{
	if (mi.order() > order)
		throw ogu::Exception(__FILE__,__LINE__,
							 "Wrong order of multi-index in FormulaDerivatives::derivative");
	if (mi.nv() != nvar)
		throw ogu::Exception(__FILE__,__LINE__,
							 "Wrong multi-index variables in FormulaDerivatives::derivative");

	int skip = ptriang.noverk(nvar+mi.order()-1, mi.order()-1);
	return d[skip+mi.offset()];
}

void FormulaDerivatives::print(const OperationTree& otree) const
{
	for (int i = 1; i < ptriang.noverk(nvar+order,order); i++) {
		printf("derivative i=%d is formula %d\n", i, d[i]);
		otree.print_operation(d[i]);
	}
}

void FormulaEvaluator::eval(const AtomValues& av, FormulaEvalLoader& loader)
{
	etree.reset_all();
	av.setValues(etree);
	for (unsigned int i = 0; i < formulas.size(); i++) {
		double res = etree.eval(formulas[i]);
		loader.load((int)i, res);
	}
}

FoldMultiIndex::FoldMultiIndex(int nv, int ordd)
	: nvar(nv), ord(ordd), data(new int[ord])
{
	for (int i = 0; i < ord; i++)
		data[i] = 0;
}

/** Note that a monotone sequence mapped by monotone mapping yields a
 * monotone sequence. */
FoldMultiIndex::FoldMultiIndex(int nv, const FoldMultiIndex& mi, const vector<int>& mp)
	: nvar(nv), ord(mi.ord), data(new int[ord])
{
	for (int i = 0; i < ord; i++) {
		if (i < ord-1 && mp[i+1] < mp[i])
			throw ogu::Exception(__FILE__,__LINE__,
								 "Mapping not monotone in FoldMultiIndex constructor");
		if (mp[mi[i]] >= nv || mp[mi[i]] < 0)
			throw ogu::Exception(__FILE__,__LINE__,
								 "Mapping out of bounds in FoldMultiIndex constructor");
		data[i] = mp[mi[i]];
	}
}

void FoldMultiIndex::increment()
{
	if (ord == 0)
		return;

	int k = ord-1;
	data[k]++;
	while (k > 0 && data[k] == nvar) {
		data[k] = 0;
		data[--k]++;
	}
	for (int kk = 1; kk < ord; kk++)
		if (data[kk-1] > data[kk])
			data[kk] = data[kk-1];
}


// For description of an algorithm for calculation of folded offset,
// see Tensor Library Documentation, Ondra Kamenik, 2005, description
// of FTensor::getOffsetRecurse().
int FoldMultiIndex::offset() const
{
	// make copy for the recursions
	int* tmp = new int[ord];
	for (int i = 0; i < ord; i++)
		tmp[i] = data[i];
	// call the recursive algorithm
	int res = offset_recurse(tmp, ord, nvar);

	delete [] tmp;
	return res;
}

int FoldMultiIndex::offset_recurse(int* data, int len, int nv)
{
	if (len == 0)
		return 0;
	// calculate length of initial constant indices
	int prefix = 1;
	while (prefix < len && data[0] == data[prefix])
		prefix++;

	int m = data[0];
	int s1 = ptriang.noverk(nv+len-1, len) - ptriang.noverk(nv-m+len-1,len);

	// cancel m from the rest of the sequence
	for (int i = prefix; i < len; i++)
		data[i] -= m;

	// calculate offset of the remaining sequence
	int s2 = offset_recurse(data+prefix, len-prefix, nv-m);
	// return the sum
	return s1+s2;
}

FormulaDerEvaluator::FormulaDerEvaluator(const FormulaParser& fp)
	: etree(fp.otree, -1)
{
	for (unsigned int i = 0; i < fp.ders.size(); i++)
		ders.push_back((const FormulaDerivatives*)(fp.ders[i]));

	der_atoms = fp.atoms.variables();
}

void FormulaDerEvaluator::eval(const AtomValues& av, FormulaDerEvalLoader& loader, int order)
{
	if (ders.size() == 0)
		return;
	int maxorder = ders[0]->order;

	if (order > maxorder)
		throw ogu::Exception(__FILE__,__LINE__,
							 "Wrong order in FormulaDerEvaluator::eval");

	etree.reset_all();
	av.setValues(etree);

	int nvar = der_atoms.size();
	int* vars = new int[order];
	int order_beg = ptriang.noverk(nvar+order-1, order-1);
	int order_end = ptriang.noverk(nvar+order, order);
	for (unsigned int i = 0; i < ders.size(); i++) {
		FoldMultiIndex mi(nvar, order);
		for (int j = order_beg; j < order_end; j++) {
			if (ders[i]->d[j] != OperationTree::zero) {
				// set vars from multiindex mi and variables
				for (int k = 0; k < order; k++)
					vars[k] = der_atoms[mi[k]];
				// evaluate
				double res = etree.eval(ders[i]->d[j]);
				// load
				loader.load(i, order, vars, res);
			}
			mi.increment();
		}
	}

	delete [] vars;
}

void FormulaDerEvaluator::eval(const vector<int>& mp, const AtomValues& av,
							   FormulaDerEvalLoader& loader, int order)
{
	etree.reset_all();
	av.setValues(etree);

	int nvar_glob = der_atoms.size();
	int nvar = mp.size();
	int* vars = new int[order];
	int order_beg = ptriang.noverk(nvar+order-1, order-1);

	for (unsigned int i = 0; i < ders.size(); i++) {
		FoldMultiIndex mi(nvar, order);
		do {
			// find index of the derivative in the tensor
			FoldMultiIndex mi_glob(nvar_glob, mi, mp);
			int j = order_beg + mi_glob.offset();
			// if it is not a zero
			if (ders[i]->d[j] != OperationTree::zero) {
				// set vars from the global multiindex
				for (int k = 0; k < order; k++)
					vars[k] = der_atoms[mi_glob[k]];
				// evaluate derivative
				double res = etree.eval(ders[i]->d[j]);
				// load
				loader.load(i, order, vars, res);
			}
			mi.increment();
		} while (! mi.past_the_end());
	}

	delete [] vars;
}


// Local Variables:
// mode:C++
// End:
