// Copyright (C) 2005, Ondra Kamenik

// $Id: formula_parser.h 762 2006-05-22 13:00:07Z kamenik $

#ifndef OGP_FORMULA_PARSER_H
#define OGP_FORMULA_PARSER_H

#include "tree.h"

namespace ogp {
	using std::vector;

	/** Pure virtual class defining a minimal interface for
	 * representation of nulary terms within FormulaParser. */
	class Atoms {
	public:
		Atoms() {}
		virtual ~Atoms() {}
		/** This returns previously assigned internal index to the
		 * given atom, or returns -1 if the atom has not been assigned
		 * yet. The method can raise an exception, if the Atoms
		 * implementation is strict and the name is not among
		 * prescribed possible values. */
		virtual int check(const char* name) const = 0;
		/** This method assigns an internal index to the nulary term
		 * described by the name. The internal index is allocated by
		 * OperationTree class. */
		virtual void assign(const char* name, int t) = 0;
		/** Returns a number of variables which will be used for
		 * differentiations. */
		virtual int nvar() const = 0;
		/** Returns a vector of variable's internal indices which will
		 * be used for differentiations. */
		virtual vector<int> variables() const = 0;
		/** Debug print. */
		virtual void print() const = 0;
	};

	/** Pure virtual class defining interface for all classes able to
	 * set nulary terms to evaluation tree EvalTree. The
	 * implementations of this class will have to be connected with
	 * Atoms to have knowledge about the atoms and their indices in
	 * the tree, and will call EvalTree::set_nulary. */
	class AtomValues {
	public:
		virtual ~AtomValues() {}
		virtual void setValues(EvalTree& et) const = 0;
	};

	class FormulaDerEvaluator;
	class FoldMultiIndex;

	/** This class represents indices of all derivatives of one
	 * formula. The derivatives include also the zero-th derivative
	 * which is the formula itself. The indices are stored in one long
	 * row implemented as an array of intergers rather than
	 * sdt::vector<int>, since it might be very long.  <p>The ordering
	 * in the row corresponds to a sequence of folded tensor indices
	 * of the maximum order with one extra variable. The extra
	 * variable (one can imagine it as X) means that no derivative is
	 * taken for this variable. For instance derivatives of maximum
	 * order of 3 of the variables 0, 1 and 2, will be ordered in the
	 * following ordering: XXX, XX0, XX1, XX2, X00, X01, X02, X11,
	 * X12, X22, 000, 001, 002, 011, 012, 022, 111, 112, 122, 222. The
	 * first XXX is zero-th derivative, i.e. the formula itself; next
	 * three are the first order derivatives; next 6 are the second
	 * order derivatives, and next 10 are the third order derivatives.
	 */
	class FormulaDerivatives {
		friend class FormulaDerEvaluator;
	protected:
		/** The array of derivative indices. */
		int* d;
		/** The number of variables. */
		int nvar;
		/** The maximum order of derivatives. */
		int order;
	public:
		/** The constructor allocates and fills the sequence of the
		 * indices of derivatives for a formula.
		 * @param otree the OperationTree for which all work is done
		 * and to which the derivatives are added.
		 * @param vars the vector of nulary terms in the tree; the
		 * derivatives are taken with respect to these variables in
		 * the ordering given by the vector.
		 * @param f the index of the formula being differentiated. d[0] is set to f.
		 * @param max_order the maximum order of differentiation.
		 */ 
		FormulaDerivatives(OperationTree& otree, const vector<int>& vars, int f, int max_order);
		/** Copy constructor. */
		FormulaDerivatives(const FormulaDerivatives& fd);
		virtual ~FormulaDerivatives()
			{ if (d) delete [] d; }
		/** Random access to the derivatives via multiindex. This
		 * should be used with care since it is not very efficient. It
		 * returns the tree index of the selected derivative. */
		int derivative(const FoldMultiIndex& mi) const;
		/** Debug print. */
		void print(const OperationTree& otree) const;
	};

	class FormulaEvaluator;

	/** This class is able to parse a number of formulas and
	 * differentiate them. The life cycle of the object is as follows:
	 * After it is created, a few calls to parse will add formulas
	 * (zero derivatives) to the object. Then a method differentiate()
	 * can be called and a vector of pointers to derivatives for each
	 * formula is created. After this, no one should call other
	 * parse() or differentiate(). A const reference of the object can
	 * be used in constructors of FormulaEvaluator and
	 * FormulaDerEvaluator in order to evaluate formulas (zero
	 * derivatives) and higher derivatives resp. */
	class FormulaParser {
		friend class FormulaEvaluator;
		friend class FormulaDerEvaluator;
	protected:
		/** The OperationTree of all formulas, including derivatives. */
		OperationTree otree;
		/** Reference to Atoms. The Atoms are filled with nulary terms
		 * during execution of parse(). */
		Atoms& atoms;
		/** Vector of formulas (zero derivatives) in the order as they
		 * have been parsed. */
		vector<int> formulas;
		/** The vector to derivatives, each vector corresponds to a
		 * formula in the vector formulas. */
		vector<FormulaDerivatives*> ders;
	public:
		/** Construct an empty formula parser. */
		FormulaParser(Atoms& a)
			: atoms(a) {}
		/** Copy constructor using a different instance of Atoms. */
		FormulaParser(const FormulaParser& fp, Atoms& a);
		virtual ~FormulaParser();

		/** Requires an addition of the formula; called from the
		 * parser. */
		void add_formula(int t);
		/** Requires an addition of the binary operation; called from
		 * the parser. */
		int add_binary(code_t code, int t1, int t2);
		/** Requires an addition of the unary operation; called from
		 * the parser. */		
		int add_unary(code_t code, int t);
		/** Requires an addition of the nulary operation given by the
		 * string. The Atoms are consulted for uniquness and are given
		 * an internal index generated by the OperationTree. This is
		 * the channel through which the Atoms are filled. */
		int add_nulary(const char* str);

		/** Parse a given string containing one or more formulas. The
		 * formulas are parsed and added to the OperationTree and to
		 * the formulas vector. */
		void parse(int length, const char* stream);
		/** Processes a syntax error from bison. */
		void error(const char* mes) const;
		/** Differentiate all the formulas up to the given order. The
		 * variables with respect to which the derivatives are taken
		 * are obtained by Atoms::variables(). */
		void differentiate(int max_order);
		/** Return i-th formula derivatives. */
		const FormulaDerivatives& derivatives(int i) const;

		/** This returns a maximum index of zero derivative
		 * formulas. This is a maximum of the vector formulas. */
		int last_formula() const;

		/** This returns a number of formulas. */
		int nformulas() const
			{return (int)(formulas.size());}

		/** This returns a reference to atoms. */
		const Atoms& getAtoms() const
			{return atoms;}
		/** This returns the tree. */
		const OperationTree getTree() const
			{return otree;}

		/** Debug print. */
		void print() const;
	private:
		/** Hide this copy constructor declaration by declaring it as
		 * private. */
		FormulaParser(const FormulaParser& fp);
	};

	/** This is a pure virtual class defining an interface for all
	 * classes which will load the results of formula (zero
	 * derivative) evaluations. A primitive implementation of this
	 * class can be a vector of doubles. */
	class FormulaEvalLoader {
	public:
		virtual ~FormulaEvalLoader() {}
		/** Set the value res for the given formula. The formula is
		 * identified by an index corresponding to the ordering in
		 * which the formulas have been parsed (starting from
		 * zero). */
		virtual void load(int i, double res) = 0;
	};

	/** This class evaluates zero derivatives of the FormulaParser. */
	class FormulaEvaluator {
	protected:
		/** The evaluation tree consisting only from zero derivative
		 * formulas. */
		EvalTree etree;
		/** The indices of formulas within the tree. */
		vector<int> formulas;
	public:
		/** Construct from FormulaParser. */
		FormulaEvaluator(const FormulaParser& fp)
			: etree(fp.otree, fp.last_formula()), formulas(fp.formulas) {}
		/** Evaluate the formulas using the given AtomValues and load
		 * the results using the given loader. */
		void eval(const AtomValues& av, FormulaEvalLoader& loader);
	};

	/** This is a pure virtual class defining an interface for all
	 * classes which will load the results of formula derivative
	 * evaluations. */
	class FormulaDerEvalLoader {
	public:
		virtual ~FormulaDerEvalLoader() {}
		/** This loads the result of the derivative of the given
		 * order. The semantics of i is the same as in
		 * FormulaEvalLoader::load. The indices of variables with
		 * respect to which the derivative was taken are stored in
		 * memory pointed by vars. These are the tree indices. */
		virtual void load(int i, int order, const int* vars, double res) = 0;
	};

	/** This class is a utility class representing the tensor
	 * multindex. It can basically increment itself, and calculate
	 * its offset in the folded tensor. */
	class FoldMultiIndex {
		int nvar;
		int ord;
		int* data;
	public:
		/** Initializes the multiindex to zeros. */
		FoldMultiIndex(int nv, int order);
		/** Makes a new multiindex of the same order applying a given
		 * mapping to the indices. The mapping is supposed to be monotone. */
		FoldMultiIndex(int nv, const FoldMultiIndex& mi, const vector<int>& mp);
		virtual ~FoldMultiIndex()
			{delete [] data;}
		/** Increment the multiindex. */
		void increment();
		/** Return offset of the multiindex in the folded tensor. */ 
		int offset() const;
		const int& operator[](int i) const
			{return data[i];}
		/** Return order of the multiindex, i.e. dimension of the
		 * tensor. */ 
		int order() const
			{return ord;}
		/** Return the number of variables. */
		int nv() const
			{return nvar;}
		/** Return the data. */
		const int* ind() const
			{return data;}
		/** Return true if the end of the tensor is reached. The
		 * result of a subsequent increment should be considered
		 * unpredictable. */
		bool past_the_end() const
			{return (ord == 0) || (data[0] == nvar);}
	private:
		static int offset_recurse(int* data, int len, int nv);
	};

	/** This class evaluates derivatives of the FormulaParser. */
	class FormulaDerEvaluator {
		/** Its own instance of EvalTree. */
		EvalTree etree;
		/** The indices of derivatives for each formula. This is a
		 * const copy FormulaParser::ders. We do not allocate nor
		 * deallocate anything here. */
		vector<const FormulaDerivatives*> ders;
		/** A copy of tree indices corresponding to atoms to with
		 * respect the derivatives were taken. */
		vector<int> der_atoms;
	public:
		/** Construct the object from FormulaParser. */
		FormulaDerEvaluator(const FormulaParser& fp);
		/** Evaluate the derivatives from the FormulaParser wrt to all
		 * atoms in variables vector at the given AtomValues. The
		 * given loader is used for output. */
		void eval(const AtomValues& av, FormulaDerEvalLoader& loader, int order);
		/** Evaluate the derivatives from the FormulaParser wrt to a
		 * selection of atoms of the atoms in der_atoms vector at the
		 * given AtomValues. The selection is given by a monotone
		 * mapping to the indices (not values) of the der_atoms. */
		void eval(const vector<int>& mp, const AtomValues& av, FormulaDerEvalLoader& loader,
				  int order);
	};
};

#endif

// Local Variables:
// mode:C++
// End:
