
#include "dynare3.h"
#include "dynare_params.h"
#include "dynare_exception.h"

#include "utils/cc/memory_file.h"
#include "utils/cc/exception.h"
#include "parser/cc/parser_exception.h"
#include "../sylv/cc/SylvException.h"
#include "../tl/cc/tl_exception.h"
#include "../kord/approximation.h"
#include "../kord/kord_exception.h"
#include "../kord/global_check.h"

#ifndef DYNVERSION
#define DYNVERSION "unknown"
#endif

/**************************************************************************************/
/*       Dynare class                                                                 */
/**************************************************************************************/

Dynare::Dynare(const char* modname, Journal& jr)
	: journal(jr), model(NULL), vcov(NULL), ysteady(NULL), md(1), dnl(NULL), denl(NULL), dsnl(NULL),
	  fe(NULL), fde(NULL)
{
	// make memory file
	ogu::MemoryFile mf(modname);
	if (mf.exists()) {
		try {
			model = new ogdyn::DynareModel(mf.base(), mf.length());
		} catch (const ogp::ParserException& pe) {
			int line;
			int col;
			mf.line_and_col(pe.offset(), line, col);
			throw DynareException(pe.message(), modname, line, col);
		}
		// if model has vcov, take it, otherwise create a new one and
		// initialize to the unit matrix
		if (model->getVcov())
			vcov = new TwoDMatrix((const TwoDMatrix&)*(model->getVcov()));
		else {
			vcov = new TwoDMatrix(nexog(), nexog());
			vcov->unit();
		}
		ysteady = new Vector(model->getAtoms().ny());
		dnl = new DynareNameList(*this);
		denl = new DynareExogNameList(*this);
		dsnl = new DynareStateNameList(*this, *dnl, *denl);
		fe = new ogp::FormulaEvaluator(model->getParser());
		fde = new ogp::FormulaDerEvaluator(model->getParser());
	} else {
		throw DynareException(__FILE__, __LINE__, string("Could not open model file ")+modname);
	}
}

Dynare::Dynare(const Dynare& dynare)
	: journal(dynare.journal), model(NULL), vcov(NULL), ysteady(NULL), md(dynare.md),
	  dnl(NULL), denl(NULL), dsnl(NULL), fe(NULL), fde(NULL)
{
	// todo: make sure that all classes have copy constructors, much work!
	model = new ogdyn::DynareModel(*(dynare.model));
	vcov = new TwoDMatrix(*(dynare.vcov));
	ysteady = new Vector(*(dynare.ysteady));
	dnl = new DynareNameList(*this);
	denl = new DynareExogNameList(*this);
	dsnl = new DynareStateNameList(*this, *dnl, *denl);
	fe = new ogp::FormulaEvaluator(model->getParser());
	fde = new ogp::FormulaDerEvaluator(model->getParser());
}

Dynare::~Dynare()
{
	if (model)
		delete model;
	if (vcov)
		delete vcov;
	if (ysteady)
		delete ysteady;
	if (dnl)
		delete dnl;
	if (dsnl)
		delete dsnl;
	if (denl)
		delete denl;
	if (fe)
		delete fe;
	if (fde)
		delete fde;
}

void Dynare::writeMat4(FILE* fd, const char* prefix) const
{
	char tmp[100];
	sprintf(tmp, "%s_vars", prefix);
	getAllEndoNames().writeMat4(fd, tmp);
	getAllEndoNames().writeMat4Indices(fd, prefix);
	sprintf(tmp, "%s_state_vars", prefix);
	getStateNames().writeMat4(fd, tmp);
	sprintf(tmp, "%s_shocks", prefix);
	getExogNames().writeMat4(fd, tmp);
	getExogNames().writeMat4Indices(fd, prefix);
	sprintf(tmp, "%s_vcov_exo", prefix);
	getVcov().writeMat4(fd, tmp);
	TwoDMatrix aux(1,1);
	sprintf(tmp, "%s_nstat", prefix);
	aux.get(0,0) = nstat();
	aux.writeMat4(fd, tmp);
	sprintf(tmp, "%s_npred", prefix);
	aux.get(0,0) = npred();
	aux.writeMat4(fd, tmp);
	sprintf(tmp, "%s_nboth", prefix);
	aux.get(0,0) = nboth();
	aux.writeMat4(fd, tmp);
	sprintf(tmp, "%s_nforw", prefix);
	aux.get(0,0) = nforw();
	aux.writeMat4(fd, tmp);
}

void Dynare::solveDeterministicSteady(Vector& steady)
{
	JournalRecordPair pa(journal);
	pa << "Non-linear solver for deterministic steady state" << endrec;
	steady = (const Vector&) model->getInit();
	DynareVectorFunction dvf(*this);
	DynareJacobian dj(*this);
	ogu::NLSolver nls(dvf, dj, 500, 1.e-13, journal);
	int iter;
	if (! nls.solve(steady, iter))
		throw DynareException(__FILE__, __LINE__,
							  "Could not obtain convergence in non-linear solver");
}

// evaluate system at given y_t=y_{t+1}=y_{t-1}, and given shocks x_t
void Dynare::evaluateSystem(Vector& out, const Vector& yy, const Vector& xx)
{
	ConstVector yym(yy, nstat(), nys());
	ConstVector yyp(yy, nstat()+npred(), nyss());
	evaluateSystem(out, yym, yy, yyp, xx);
}

// evaluate system at given y^*_{t-1}, y_t, y^{**}_{t+1} and at
// exogenous x_t, all three vectors yym, yy, and yyp have the
// respective lengths of y^*_{t-1}, y_t, y^{**}_{t+1}
void Dynare::evaluateSystem(Vector& out, const Vector& yym, const Vector& yy,
							const Vector& yyp, const Vector& xx)
{
	DynareAtomValues dav(model->getAtoms(), model->getParams(), yym, yy, yyp, xx);
	DynareEvalLoader del(model->getAtoms(), out);
	fe->eval(dav, del);
}

void Dynare::calcDerivatives(const Vector& yy, const Vector& xx)
{
	Vector yym(yy, nstat(), nys());
	Vector yyp(yy, nstat()+npred(), nyss());
	DynareAtomValues dav(model->getAtoms(), model->getParams(), yym, yy, yyp, xx);
	DynareDerEvalLoader ddel(model->getAtoms(), md, model->getOrder());
	for (int iord = 1; iord <= model->getOrder(); iord++)
		fde->eval(dav, ddel, iord);
}

void Dynare::calcDerivativesAtSteady()
{
	Vector xx(nexog());
	xx.zeros();
	calcDerivatives(*ysteady, xx);
}

DynareNameList::DynareNameList(const Dynare& dynare)
{
	for (int i = 0; i < dynare.ny(); i++) {
		int j = dynare.model->getAtoms().y2outer_endo()[i];
		const char* name = dynare.model->getAtoms().get_endovars()[j];
		names.push_back(name);
	}
}

DynareStateNameList::DynareStateNameList(const Dynare& dynare, const DynareNameList& dnl,
										 const DynareExogNameList& denl)
{
	for (int i = 0; i < dynare.nys(); i++)
		names.push_back(dnl.getName(i+dynare.nstat()));
	for (int i = 0; i < dynare.nexog(); i++)
		names.push_back(denl.getName(i));
}

DynareExogNameList::DynareExogNameList(const Dynare& dynare)
{
	for (int i = 0; i < dynare.nexog(); i++) {
		int j = dynare.model->getAtoms().y2outer_exo()[i];
		const char* name = dynare.model->getAtoms().get_exovars()[j];
		names.push_back(name);
	}
}


void DynareAtomValues::setValues(ogp::EvalTree& et) const
{
	// set constants
	atoms.setValues(et);

	// set parameteres
	for (unsigned int i = 0; i < atoms.get_params().size(); i++) {
		int t = atoms.index(atoms.get_params()[i], 0);
		// the parameter may not appear in the tree
		if (t != -1)
			et.set_nulary(t, paramvals[i]);
	}

	// set endogenous
	for (unsigned int outer_i = 0; outer_i < atoms.get_endovars().size(); outer_i++) {
		try {
			const ogp::DynamicAtoms::Tlagmap& lmap = atoms.lagmap(atoms.get_endovars()[outer_i]);
			for (ogp::DynamicAtoms::Tlagmap::const_iterator it = lmap.begin();
				 it != lmap.end(); ++it) {
				int ll = (*it).first;
				int t = (*it).second;
				int i = atoms.outer2y_endo()[outer_i];
				if (ll == -1) {
					et.set_nulary(t, yym[i-atoms.nstat()]);
				}
				else if (ll == 0)
					et.set_nulary(t, yy[i]);
				else
					et.set_nulary(t, yyp[i-atoms.nstat()-atoms.npred()]);
			}
		} catch (const ogu::Exception& e) {
			// ignore non-referenced endogenous variables
		}
	}

	// set exogenous
	for (unsigned int outer_i = 0; outer_i < atoms.get_exovars().size(); outer_i++) {
		try {
			const ogp::DynamicAtoms::Tlagmap& lmap = atoms.lagmap(atoms.get_exovars()[outer_i]);
			for (ogp::DynamicAtoms::Tlagmap::const_iterator it = lmap.begin();
				 it != lmap.end(); ++it) {
				int ll = (*it).first;
				if (ll == 0) { // this is always true because of checks
					int t = (*it).second;
					int i = atoms.outer2y_exo()[outer_i];			
					et.set_nulary(t, xx[i]);
				}
			}
		} catch (const ogu::Exception& e) {
			// ignore non-referenced variables
		}
	}
}

DynareEvalLoader::DynareEvalLoader(const ogp::FineAtoms& a, Vector& out)
	: Vector(out)
{
	if (a.ny() != out.length())
		throw DynareException(__FILE__, __LINE__, "Wrong length of out vector in DynareEvalLoader constructor");
}

/** This clears the container of model derivatives and initializes it
 * inserting empty sparse tensors up to the given order. */
DynareDerEvalLoader::DynareDerEvalLoader(const ogp::FineAtoms& a,
										 TensorContainer<FSSparseTensor>& mod_ders,
										 int order)
	: atoms(a), md(mod_ders)
{
	md.clear();
	for (int iord = 1; iord <= order; iord++) {
		FSSparseTensor* t = new FSSparseTensor(iord, atoms.ny()+atoms.nys()+atoms.nyss()+atoms.nexo(), atoms.ny());
		md.insert(t);
	}
}

void DynareDerEvalLoader::load(int i, int iord, const int* vars, double res)
{
	FSSparseTensor* t = md.get(Symmetry(iord));
	IntSequence s(iord, 0);
	for (int j = 0; j < iord; j++)
		s[j] = atoms.get_pos_of_all(vars[j]);
	t->insert(s, i, res);
}

DynareJacobian::DynareJacobian(Dynare& dyn)
	: Jacobian(dyn.ny()), d(dyn)
{
	zeros();
}

void DynareJacobian::eval(const Vector& yy)
{
	Vector yym(yy, d.nstat(), d.nys());
	Vector yyp(yy, d.nstat()+d.npred(), d.nyss());
	Vector xx(d.nexog());
	xx.zeros();
	DynareAtomValues dav(d.getModel().getAtoms(), d.getModel().getParams(), yym, yy, yyp, xx);
	zeros();
	d.fde->eval(dav, *this, 1);
}

void DynareJacobian::load(int i, int iord, const int* vars, double res)
{
	if (iord != 1)
		throw DynareException(__FILE__, __LINE__,
							  "Derivative order different from order=1 in DynareJacobian::load");

	int t = vars[0];
	int j = d.getModel().getAtoms().get_pos_of_all(t);
	if (j < d.nyss())
		get(i, j+d.nstat()+d.npred()) += res;
	else if (j < d.nyss()+d.ny())
		get(i, j-d.nyss()) += res;
	else if (j < d.nyss()+d.ny()+d.nys())
		get(i, j-d.nyss()-d.ny()+d.nstat()) += res;
}

void DynareVectorFunction::eval(const ConstVector& in, Vector& out)
{
	check_for_eval(in, out);
	Vector xx(d.nexog());
	xx.zeros();
	d.evaluateSystem(out, in, xx);
}

int main(int argc, char** argv)
{
	DynareParams params(argc, argv);
	if (params.help) {
		params.printHelp();
		return 0;
	}
	if (params.version) {
		printf("Dynare++ v. %s. Copyright (C) 2004,2005,2006 Ondra Kamenik\n",
			   DYNVERSION);
		printf("Dynare++ comes with ABSOLUTELY NO WARRANTY and is distributed under\n");
		printf("GPL: modules integ, tl, kord, sylv, src, extern and documentation\n");
		printf("LGPL: modules parser, utils\n");
		printf(" for GPL  see http://www.gnu.org/licenses/gpl.html\n");
		printf(" for LGPL see http://www.gnu.org/licenses/lgpl.html\n");
		return 0;
	}
	THREAD_GROUP::max_parallel_threads = params.num_threads;

	try {
		// make journal name
		char jname[500];
		sprintf(jname, "%s.jnl", params.basename);
		Journal journal(jname);

		// make dynare object
		Dynare dynare(params.modname, journal);

		RandomGenerator::initSeed(params.seed);

		tls.init(dynare.order(),
				 dynare.nstat()+2*dynare.npred()+3*dynare.nboth()+
				 2*dynare.nforw()+dynare.nexog());

		// write information on variables to journal
		{
			JournalRecord jr(journal);
			jr << "The model has " << dynare.nstat() << " static, " << dynare.nboth()+dynare.npred()
			   << " predetermined, " << dynare.nboth()+dynare.nforw() << " frwd-looking, including "
			   << dynare.nboth() << " at both times" << endrec;
		}

		Approximation app(dynare, journal, params.num_steps);
		app.walkStochSteady();


		// open mat file
		char matfile[100];
		sprintf(matfile, "%s.mat", params.basename);
		FILE* matfd;
		if (NULL == (matfd=fopen(matfile, "wb"))) {
			fprintf(stderr, "Couldn't open %s for writing.\n", matfile);
			exit(1);
		}

		dynare.writeMat4(matfd, params.prefix);

		char ss_matrix_name[100];
		sprintf(ss_matrix_name, "%s_steady_states", params.prefix);
		ConstTwoDMatrix(app.getSS()).writeMat4(matfd, ss_matrix_name);

		// check the approximation
		if (params.check_along_path || params.check_along_shocks
			|| params.check_on_ellipse) {
			GlobalChecker gcheck(app, THREAD_GROUP::max_parallel_threads, journal);
			if (params.check_along_shocks)
				gcheck.checkAlongShocksAndSave(matfd, params.prefix,
											   params.getCheckShockPoints(),
											   params.getCheckShockScale(),
											   params.check_evals);
			if (params.check_on_ellipse)
				gcheck.checkOnEllipseAndSave(matfd, params.prefix,
											 params.getCheckEllipsePoints(),
											 params.getCheckEllipseScale(),
											 params.check_evals);
			if (params.check_along_path)
				gcheck.checkAlongSimulationAndSave(matfd, params.prefix,
												   params.getCheckPathPoints(),
												   params.check_evals);
		}

		// write the folded decision rule to the Mat-4 file
		app.getFoldDecisionRule().writeMat4(matfd, params.prefix);

		// simulate
		const DecisionRule& dr = app.getUnfoldDecisionRule();
		SimResultsStats res(dynare.numeq(), params.num_per);
		RandomShockRealization sr(dynare.getVcov());
		res.simulate(params.num_sim, dr, dynare.getSteady(), sr, journal);
		res.writeMat4(matfd, params.prefix);

		// impulse response functions
		if (params.do_irfs) {
			IRFResults irf(dynare, dr, res, journal);
			irf.writeMat4(matfd, params.prefix);
		}

		fclose(matfd);

	} catch (const KordException& e) {
		printf("Caugth Kord exception: ");
		e.print();
		return e.code();
	} catch (const TLException& e) {
		printf("Caugth TL exception: ");
		e.print();
		return 255;
	} catch (SylvException& e) {
		printf("Caught Sylv exception: ");
		e.printMessage();
		return 255;
	} catch (const DynareException& e) {
		printf("Caught Dynare exception: %s\n", e.message());
		return 255;
	} catch (const ogu::Exception& e) {
		printf("Caught ogu::Exception: ");
		e.print();
		return 255;
	} catch (const ogp::ParserException& e) {
		printf("Caught parser exception: %s\n", e.message());
		return 255;
	}


  return 0;
}
