// Copyright (C) 2006, Ondra Kamenik

// $Id: dynare_model.cpp 786 2006-06-13 11:58:01Z kamenik $

#include "parser/cc/parser_exception.h"
#include "parser/cc/location.h"
#include "dynare_model.h"
#include "dynare_exception.h"

#include <stdlib.h>

using namespace ogdyn;

ParsedMatrix::ParsedMatrix(const ogp::MatrixParser& mp)
	: TwoDMatrix(mp.nrows(), mp.ncols())
{
	zeros();
	for (ogp::MPIterator it = mp.begin(); it != mp.end(); ++it)
		get(it.row(), it.col()) = *it;
}

AtomAssignings::AtomAssignings(const AtomAssignings& aa)
	: atoms(aa.atoms), expr(aa.expr, atoms), left_names(aa.left_names),
	  order(aa.order)
{
	// fill the lname2expr
	for (Tvarmap::const_iterator it = aa.lname2expr.begin();
		 it != aa.lname2expr.end(); ++it)
		lname2expr.insert(Tvarmap::value_type(left_names.query((*it).first), (*it).second));
}

/** A global symbol for passing info to the AtomAssignings from
 * asgn_parse(). */
AtomAssignings* aparser;

/** The declaration of functions defined in asgn_ll.cc and asgn_tab.cc
 * generated from assign.lex assign.y */
void* asgn__scan_buffer(char*, unsigned int);
void asgn__destroy_buffer(void*);
void asgn_parse();
extern ogp::location_type asgn_lloc;

void AtomAssignings::parse(int length, const char* stream)
{
	char* buffer = new char[length+2];
	strncpy(buffer, stream, length);
	buffer[length] = '\0';
	buffer[length+1] = '\0';
	asgn_lloc.off = 0;
	asgn_lloc.ll = 0;
	void* p = asgn__scan_buffer(buffer, (unsigned int)length+2);
	aparser = this;
	asgn_parse();
	delete [] buffer;
	asgn__destroy_buffer(p);
}

void AtomAssignings::error(const char* mes)
{
	throw ogp::ParserException(mes, asgn_lloc.off);
}

void AtomAssignings::add_assignment(int asgn_off, const char* str, int name_len,
									int right_off, int right_len)
{
	// the order of doing things here is important: since the
	// FormulaParser requires that all references from the i-th tree
	// refere to trees with index lass than i, so to capture also a
	// nulary term for the left hand side, it must be inserted to the
	// expression tree before the expression is parsed.

	// find the name in the atoms, make copy of name to be able to put
	// '\0' at the end
	char* buf = new char[name_len+1];
	strncpy(buf, str, name_len);
	buf[name_len] = '\0';
	// if left hand side is a registered atom, insert it to tree
	int t;
	try {
		if (atoms.check(buf))
			t = expr.add_nulary(buf);
		else
			t = -1;
	} catch (const ogp::ParserException& e) {
		t = -1;
	}
	// register left hand side in order
	order.push_back(t);

	// parse expression on the right
	try {
		expr.parse(right_len, str+right_off);
	} catch (const ogp::ParserException& e) {
		throw ogp::ParserException(e, asgn_off+right_off);
	}

	// register name of the left hand side and put to lname2expr
	const char* ss = left_names.insert(buf);
	lname2expr.insert(Tvarmap::value_type(ss, order.size()-1));

	// delete name
	delete [] buf;
}


void AtomAssignings::print() const
{
	printf("atoms\n");
	atoms.print();
}

void AtomAsgnEvaluator::setValues(ogp::EvalTree& et) const
{
	// set values of constants
	aa.atoms.setValues(et);

	// set values of variables to NaN or to user set values
	double nan = strtod("NaN",NULL);
	for (int i = 0; i < aa.atoms.nvar(); i++) {
		const char* ss = aa.atoms.name(i);
		int t = aa.atoms.index(ss);
		if (t >= 0) {
			Tusrvalmap::const_iterator it = user_values.find(t);
			if (it == user_values.end())
				et.set_nulary(t, nan);
			else
				et.set_nulary(t, (*it).second);
		}
	}
}

void AtomAsgnEvaluator::set_user_value(const char* name, double val)
{
	int t = aa.atoms.index(name);
	if (t >= 0) {
		Tusrvalmap::iterator it = user_values.find(t);
		if (it == user_values.end())
			user_values.insert(Tusrvalmap::value_type(t, val));
		else
			(*it).second = val;
	}
}

void AtomAsgnEvaluator::load(int i, double res)
{
	// set the value
	operator[](i) = res;
	// if i-th expression is atom, set its value to this EvalTree
	int t = aa.order[i];
	if (t >= 0)
		etree.set_nulary(t, res);
}

double AtomAsgnEvaluator::get_value(const char* name) const
{
	AtomAssignings::Tvarmap::const_iterator it = aa.lname2expr.find(name);
	if (it == aa.lname2expr.end())
		return strtod("NaN",NULL);
	else
		return operator[]((*it).second);
}

extern ogp::location_type dynglob_lloc;

DynareModel::DynareModel(const char* stream, int len)
	: atoms(), eqs(atoms),
	  paramset(), initval(),
	  param_vals(0), init_vals(0),
	  vcov_mat(0),
	  model_beg(0), model_end(-1),
	  paramset_beg(0), paramset_end(-1),
	  initval_beg(0), initval_end(-1),
	  vcov_beg(0), vcov_end(-1),
	  order_beg(0), order_end(-1),
	  order(-1)
{
	try {
		parse_glob(len, stream);
	} catch (const ogp::ParserException& e) {
		throw ogp::ParserException(e, dynglob_lloc.off);
	}
	try {
		if (paramset_end > paramset_beg)
			paramset.parse(paramset_end-paramset_beg, stream+paramset_beg);
	} catch (const ogp::ParserException& e) {
		throw ogp::ParserException(e, paramset_beg);
	}
	try {
		if (model_end > model_beg)
			eqs.parse(model_end-model_beg, stream+model_beg);
		else
			throw ogp::ParserException("Model section not found.", 0);
	} catch (const ogp::ParserException& e) {
		throw ogp::ParserException(e, model_beg);
	}
	try {
		if (initval_end > initval_beg)
			initval.parse(initval_end-initval_beg, stream+initval_beg);
	} catch (const ogp::ParserException& e) {
		throw ogp::ParserException(e, initval_beg);
	}
	try {
		if (vcov_end > vcov_beg) {
			vcov.parse(vcov_end-vcov_beg, stream+vcov_beg);
			vcov_mat = new ParsedMatrix(vcov);
		}
	} catch (const ogp::ParserException& e) {
		throw ogp::ParserException(e, vcov_beg);
	}
	if (order_end > order_beg) {
		char* buf = new char[order_end-order_beg+1];
		strncpy(buf, stream+order_beg, order_end-order_beg);
		buf[order_end-order_beg] = '\0';
		sscanf(buf, "%d", &order);
		delete [] buf;
	}
	atoms.parsing_finished(ogp::VarOrdering::bfspbfpb);
	calc_params();
	calc_init();
	check_model();
	if (order >= 1)
		eqs.differentiate(order);
}

DynareModel::DynareModel(const DynareModel& dm)
	: atoms(dm.atoms), eqs(dm.eqs, atoms), paramset(dm.paramset), initval(dm.initval),
	  vcov(dm.vcov), param_vals(NULL), init_vals(NULL), vcov_mat(NULL),
	  model_beg(dm.model_beg), model_end(dm.model_end),
	  paramset_beg(dm.paramset_beg), paramset_end(dm.paramset_end),
	  initval_beg(dm.initval_beg), initval_end(dm.initval_end),
	  vcov_beg(dm.vcov_beg), vcov_end(dm.vcov_end),
	  order_beg(dm.order_beg), order_end(dm.order_end),
	  order(dm.order)
{
	if (dm.param_vals)
		param_vals = new Vector((const Vector&)*(dm.param_vals));
	if (dm.init_vals)
		init_vals = new Vector((const Vector&)*(dm.init_vals));
	if (dm.vcov_mat)
		vcov_mat = new TwoDMatrix((const TwoDMatrix&)*(dm.vcov_mat));
}

DynareModel::~DynareModel()
{
	if (param_vals)
		delete param_vals;
	if (init_vals)
		delete init_vals;
	if (vcov_mat)
		delete vcov_mat;
}

void DynareModel::add_name(const char* name, int flag)
{
	if (flag == 1) {
		// endogenous
		atoms.register_uniq_endo(name);
		initval.register_uniq(name);
	} else if (flag == 2) {
		// exogenous
		atoms.register_uniq_exo(name);
	} else if (flag == 3) {
		// parameter
		atoms.register_uniq_param(name);
		paramset.register_uniq(name);
		initval.register_uniq(name);
	} else {
		throw DynareException(__FILE__, __LINE__,
							  "Unrecognized flag value.");
	}
}

/** A global symbol for passing info to the DynareModel from
 * parser. */
DynareModel* dynare_parser;

/** The declarations of functions defined in dynglob_ll.cc and
 * dynglob_tab.cc generated from dynglob.lex and dynglob.y */
void* dynglob__scan_buffer(char*, unsigned int);
void dynglob__destroy_buffer(void*);
void dynglob_parse();
extern ogp::location_type dynglob_lloc;

void DynareModel::parse_glob(int length, const char* stream)
{
	char* buffer = new char[length+2];
	strncpy(buffer, stream, length);
	buffer[length] = '\0';
	buffer[length+1] = '\0';
	void* p = dynglob__scan_buffer(buffer, (unsigned int)length+2);
	dynare_parser = this;
	dynglob_parse();
	delete [] buffer;
	dynglob__destroy_buffer(p);
}

void DynareModel::calc_params()
{
	if (param_vals)
		delete param_vals;

	param_vals = new Vector(atoms.np());
	AtomAsgnEvaluator aae(paramset);
	aae.eval();
	for (int i = 0; i < atoms.np(); i++)
		(*param_vals)[i] = aae.get_value(atoms.get_params()[i]);
}

void DynareModel::calc_init()
{
	if (init_vals)
		delete init_vals;

	init_vals = new Vector(atoms.ny());
	AtomAsgnEvaluator aae(initval);
	for (int ip = 0; ip < atoms.np(); ip++)
		aae.set_user_value(atoms.get_params()[ip], (*param_vals)[ip]);
	aae.eval();
	for (int outer = 0; outer < atoms.ny(); outer++) {
		int i = atoms.outer2y_endo()[outer];
		(*init_vals)[i] = aae.get_value(atoms.get_endovars()[outer]);
	}
}

void DynareModel::check_model() const
{
	if (atoms.ny() != eqs.nformulas()) {
		char mes[1000];
		sprintf(mes, "Model has %d equations for %d endogenous variables", eqs.nformulas(), atoms.ny());
		throw DynareException(__FILE__, __LINE__, mes);
	}
	
	int mlag, mlead;
	atoms.exovarspan(mlead, mlag);
	if (atoms.nexo() > 0 && (mlead != 0 || mlag != 0))
		throw DynareException(__FILE__,__LINE__,
							  "The model contains occurrences of lagged/leaded exogenous variables");

	atoms.endovarspan(mlead, mlag);
	if (mlead > 1 || mlag < -1)
		throw DynareException(__FILE__,__LINE__,
							  "The model contains occurrences of too lagged/leaded endogenous variables");
}

void DynareModel::error(const char* mes)
{
	// throwing zero offset since this exception will be caugth at
	// constructor
	throw ogp::ParserException(mes, 0);
}

void DynareModel::print() const
{
	printf("all atoms:\n");
	atoms.print();
	printf("parameter atoms:\n");
	paramset.print();
	printf("initval atoms:\n");
	initval.print();
	printf("model position: %d %d\n", model_beg, model_end);
	printf("paramset position: %d %d\n", paramset_beg, paramset_end);
	printf("initval position: %d %d\n", initval_beg, initval_end);
	eqs.print();
}
