// Copyright (C) 2005, Ondra Kamenik

// $Id: dynare_model.h 762 2006-05-22 13:00:07Z kamenik $

#ifndef OGDYN_DYNARE_MODEL
#define OGDYN_DYNARE_MODEL

#include "parser/cc/matrix_parser.h"

#include "dynare_atoms.h"
#include "twod_matrix.h"

#include "Vector.h"
#include "GeneralMatrix.h"

#include <map>

namespace ogdyn {
	using std::map;

	/** This represents an interval in a string by the pair of
	 * positions (including the first, excluding the second). A
	 * position is given by the line and the column within the line
	 * (both starting from 1). */
	struct PosInterval {
		int fl;
		int fc;
		int ll;
		int lc;
		PosInterval() {}
		PosInterval(int ifl, int ifc, int ill, int ilc)
			: fl(ifl), fc(ifc), ll(ill), lc(ilc) {}
		const PosInterval& operator=(const PosInterval& pi)
			{fl = pi.fl; fc = pi.fc; ll = pi.ll; lc = pi.lc; return *this;}
		/** This returns the interval beginning and interval length
		 * within the given string. */
		void translate(const char* beg, int len, const char*& ibeg, int& ilen) const;
		/** Debug print. */
		void print() const
			{printf("fl=%d fc=%d ll=%d lc=%d\n",fl,fc,ll,lc);}
	};

	/** This class is basically a GeneralMatrix but is created from
	 * parsed matrix data. */
	class ParsedMatrix : public TwoDMatrix {
	public:
		/** Construct the object from the parsed data of ogp::MatrixParser. */
		ParsedMatrix(const ogp::MatrixParser& mp);
	};

	class AtomAsgnEvaluator;

	/** This class represents atom assignments used in parameters
	 * settings and initval initialization. It maintains atoms of the
	 * all expressions on the right hand side, the parsed formulas of
	 * the right hand sides, and the information about the left hand
	 * sides. See documentation to the order member below. */
	class AtomAssignings {
		friend class AtomAsgnEvaluator;
	protected:
		typedef map<const char*, int, ogp::ltstr> Tvarmap;
		/** All atoms which should be sufficient for formulas at the
		 * right hand sides. Since the atoms of type
		 * DynareStaticAtoms, the parser will not accept variables
		 * which are not preregistered. */
		DynareStaticAtoms atoms;
		/** The formulas of right hand sides. */
		ogp::FormulaParser expr;
		/** Name storage of the names from left hand sides. */
		ogp::NameStorage left_names;
		/** Information on left hand sides. This maps a name to the
		 * index of its assigned exrpession in expr. */
		Tvarmap lname2expr;
		/** Information on left hand sides. If order[i] >= 0, then it
		 * says that i-th expression in expr is assigned to atom with
		 * order[i] tree index. */
		std::vector<int> order;
	public:
		AtomAssignings(const AtomAssignings& aa);
		AtomAssignings() : expr(atoms)
			{}
		virtual ~AtomAssignings()
			{}
		/** Parse the assignments from the given string. */
		void parse(int length, const char* stream);
		/** Process a syntax error from bison. */
		void error(const char* mes);
		/** Add an assignment. Called from assign.y. */
		void add_assignment(int asgn_off, const char* str, int name_len,
							int right_off, int right_len);
		/** This adds a variable to atoms. This is called from a
		 * global parser before the assignments are parsed. */
		void register_uniq(const char* name)
			{atoms.register_uniq(name);}
		/** Debug print. */
		void print() const;
	};

	/** This class basically evaluates the atom assignments
	 * AtomAssignings, so it inherits from ogp::FormulaEvaluator. It
	 * is also a storage for the results of the evaluation stored as a
	 * vector, so the class inherits from DVector and
	 * ogp::FormulaEvalLoader. As the expressions for atoms are
	 * evaluated, the results are values for atoms which will be
	 * used in subsequent evaluations. For this reason, the class
	 * inherits also from AtomValues. */
	class AtomAsgnEvaluator : public ogp::FormulaEvalLoader,
							  public ogp::AtomValues,
							  protected ogp::FormulaEvaluator,
							  public Vector {
	protected:
		typedef std::map<int, double> Tusrvalmap;
		Tusrvalmap user_values;
		const AtomAssignings& aa;
	public:
		AtomAsgnEvaluator(const AtomAssignings& a)
			: FormulaEvaluator(a.expr), Vector(a.expr.nformulas()), aa(a) {}
		virtual ~AtomAsgnEvaluator() {}
		/** This sets all initial values to NaNs, all constants and
		 * all values set by user by call set_value. This is called by
		 * FormulaEvaluator::eval() method, which is called by eval()
		 * method passing this argument as AtomValues. So the
		 * ogp::EvalTree will be always this->etree. */
		void setValues(ogp::EvalTree& et) const;
		/** User setting of the values. For example in initval,
		 * parameters are known and should be set to their values. In
		 * constrast endogenous variables are set initially to NaNs by
		 * AtomValues::setValues. */
		void set_user_value(const char* name, double val);
		/** This sets the result of i-th expression in aa to res, and
		 * also checks whether the i-th expression is an atom. If so,
		 * it sets the value of the atom in ogp::EvalTree
		 * this->etree. */
		void load(int i, double res);
		/** After the user values have been set, the assignments can
		 * be evaluated. For this purpose we have eval() method. The
		 * result is that this object as DVector will contain the
		 * values. It is ordered given by formulas in expr. */
		void eval()
			{ogp::FormulaEvaluator::eval(*this, *this);}
		/** This returns a value for a given name. If the name is not
		 * found among atoms, or there is no assignment for the atom,
		 * NaN is returned. */
		double get_value(const char* name) const;
	};

	class DynareModel {
	protected:
		/** All atoms for whole model. */
		DynareDynamicAtoms atoms;
		/** Parsed model equations. */
		ogp::FormulaParser eqs;
		/** Assignments for the parameters. */
		AtomAssignings paramset;
		/** Assignments for the initval. */
		AtomAssignings initval;
		/** Matrix parser for vcov. */
		ogp::MatrixParser vcov;
		/** A vector of parameters values created by calc_params. It
		 * is stored with natural ordering (outer) of the parameters
		 * given by atoms. */
		Vector* param_vals;
		/** A vector of initial values created by calc_init. It is
		 * stored with internal ordering given by atoms. */
		Vector* init_vals;
		/** A matrix for vcov. It is created after parsing of vcov is finished. */
		TwoDMatrix* vcov_mat;
	public:
		DynareModel(const char* str, int len);
		DynareModel(const DynareModel& dm);
		virtual ~DynareModel();
		const DynareDynamicAtoms& getAtoms() const
			{return atoms;}
		const ogp::FormulaParser& getParser() const
			{return eqs;}
		int getOrder() const
			{return order;}
		/** Adds a name of endogenous, exogenous or a parameter. The
		 * sort is governed by the flag. See dynglob.y for values of
		 * the flag. */
		void add_name(const char* name, int flag);
		/** Sets position of the model section. Called from
		 * dynglob.y. */
		void set_model_pos(int off1, int off2)
			{model_beg = off1; model_end = off2;}
		/** Sets position of the section setting parameters. Called
		 * from dynglob.y. */
		void set_paramset_pos(int off1, int off2)
			{paramset_beg = off1; paramset_end = off2;}
		/** Sets position of the initval section. Called from
		 * dynglob.y. */
		void set_initval_pos(int off1, int off2)
			{initval_beg = off1; initval_end = off2;}
		/** Sets position of the vcov section. Called from
		 * dynglob.y. */
		void set_vcov_pos(int off1, int off2)
			{vcov_beg = off1; vcov_end = off2;}
		/** Parser the given string as integer and set to as the
		 * order. */
		void set_order_pos(int off1, int off2)
			{order_beg = off1; order_end = off2;}
		/** Processes a syntax error from bison. */
		void error(const char* mes);
		/** Debug print. */
		void print() const;
		/** Return the vector of parameter values. */
		const Vector& getParams() const
			{return *param_vals;}
		/** Return the vector of initial values of endo variables. */
		const Vector& getInit() const
			{return *init_vals;} 
		/** Return the vcov matrix. If the vcov section has not been
		 * asserted, returns null. */
		const TwoDMatrix* getVcov() const
			{return vcov_mat;}
	protected:
		void parse_glob(int length, const char* stream);
		void calc_params();
		void calc_init();
		/** This checks the model consistency. Thuis includes: number
		 * of endo variables and number of equations, min and max lag
		 * of endogenous variables and occurrrences of exogenous
		 * variables. It throws an exception, if there is a problem. */
		void check_model() const;
	private:
		int model_beg, model_end;
		int paramset_beg, paramset_end;
		int initval_beg, initval_end;
		int vcov_beg, vcov_end;
		int order_beg, order_end;
		int order;
	};

};

#endif

// Local Variables:
// mode:C++
// End:
